#!/usr/bin/env python
# -*- coding: utf-8 -*-
# vim: sw=4:ts=4:expandtab
"""
mezmorize.utils
~~~~~~~~~~~~~~~

Provides mezmorize utility functions
"""

from copy import copy
from functools import wraps
from os import getenv
from os import path as p
from pathlib import Path
from pickle import DEFAULT_PROTOCOL
from subprocess import DEVNULL, call
from typing import Optional

try:
    import pylibmc
except ImportError:
    pylibmc = None

try:
    import pymemcache
except ImportError:
    pymemcache = None

try:
    import bmemcached
except ImportError:
    bmemcached = None

try:
    import redis
except ImportError:
    redis = None

BASEDIR = p.dirname(__file__)
DEF_CACHE_DIR = Path(BASEDIR).joinpath("cache", f"pickle_v{DEFAULT_PROTOCOL}")

DEF_THRESHOLD = 500
DEF_DEFAULT_TIMEOUT = 300
DEF_MC_HOST = DEF_REDIS_HOST = "localhost"
DEF_MC_PORT = 11211
DEF_REDIS_PORT = 6379
ENCODING = "utf-8"

ALL_MEMCACHES = (
    ("pylibmc", pylibmc),
    ("bmemcached", bmemcached),
    ("pymemcache", pymemcache),
)

DEF_MC_SERVERS = "{}:{}".format(DEF_MC_HOST, DEF_MC_PORT)
MC_SERVERS = getenv("MEMCACHIER_SERVERS") or getenv("MEMCACHEDCLOUD_SERVERS")
MC_SERVERS = MC_SERVERS or getenv("MEMCACHE_SERVERS") or DEF_MC_SERVERS
MC_USERNAME = getenv("MEMCACHIER_USERNAME") or getenv("MEMCACHEDCLOUD_USERNAME")
MC_PASSWORD = getenv("MEMCACHIER_PASSWORD") or getenv("MEMCACHEDCLOUD_PASSWORD")

REDIS_HOST = getenv("REDIS_PORT_6379_TCP_ADDR", DEF_REDIS_HOST)
DEF_REDIS_URL = "redis://{}:{}".format(REDIS_HOST, DEF_REDIS_PORT)
REDIS_URL = getenv("REDIS_URL") or getenv("REDISTOGO_URL") or DEF_REDIS_URL

CACHE_DIR = getenv("CACHE_DIR", default=DEF_CACHE_DIR)

CACHE_CONFIGS = {
    "simple": {"CACHE_TYPE": "simple"},
    "null": {"CACHE_TYPE": "null"},
    "redis": {"CACHE_TYPE": "redis", "CACHE_REDIS_URL": REDIS_URL},
    "filesystem": {"CACHE_TYPE": "filesystem", "CACHE_DIR": CACHE_DIR},
    "memcached": {"CACHE_TYPE": "memcached", "CACHE_MEMCACHED_SERVERS": [MC_SERVERS]},
    "saslmemcached": {
        "CACHE_TYPE": "saslmemcached",
        "CACHE_MEMCACHED_SERVERS": [MC_SERVERS],
        "CACHE_MEMCACHED_USERNAME": MC_USERNAME,
        "CACHE_MEMCACHED_PASSWORD": MC_PASSWORD,
    },
    "spreadsaslmemcached": {
        "CACHE_TYPE": "spreadsaslmemcached",
        "CACHE_MEMCACHED_SERVERS": [MC_SERVERS],
        "CACHE_MEMCACHED_USERNAME": MC_USERNAME,
        "CACHE_MEMCACHED_PASSWORD": MC_PASSWORD,
    },
}

HEROKU = getenv("HEROKU") or (getenv("DYNO") and getenv("HOME") == "/app")

HEROKU_PROCESSES = {
    "postgres": ["DATABASE_URL"],
    "redis": ["REDIS_URL", "REDISTOGO_URL"],
    "memcache": ["MEMCACHIER_SERVERS", "MEMCACHE_SERVERS"],
}


def pgrep(process):
    envs = HEROKU_PROCESSES.get(process, [])
    any_env = HEROKU and any(map(getenv, envs))
    return any_env or call(["pgrep", process], stdout=DEVNULL) == 0


HAS_MEMCACHE = (pylibmc or pymemcache or bmemcached) and pgrep("memcache")
AVAIL_MEMCACHES = [k for k, v in ALL_MEMCACHES if HAS_MEMCACHE and v]
HAS_REDIS = redis and pgrep("redis")


def parse_verbosity(verbose=0, quiet: Optional[bool] = None):
    if quiet:
        verbosity = "0"
    elif verbose:
        verbosity = str(verbose)
    else:
        verbosity = ""

    return verbosity


def get_cache_type(cache=None, spread=False, cache_dir=CACHE_DIR, **kwargs):
    if HAS_REDIS and HAS_MEMCACHE and not cache:
        cache = "memcached"
    elif not cache:
        cache = "redis" if HAS_REDIS else "memcached"

    if HAS_MEMCACHE and "memcached" in cache:
        if MC_USERNAME and spread:
            cache_type = "spreadsaslmemcached"
        elif MC_USERNAME:
            cache_type = "saslmemcached"
        else:
            cache_type = "memcached"
    elif HAS_REDIS and cache == "redis":
        cache_type = "redis"
    elif cache_dir and cache not in {"simple", "null"}:
        cache_type = "filesystem"
    elif cache != "null":
        cache_type = "simple"
    else:
        cache_type = "null"

    return cache_type


def get_cache_config(cache_type, db=None, SELF_SIGNED_CERTS=False, **kwargs):
    config = copy(CACHE_CONFIGS[cache_type])
    redis_url = config.get("CACHE_REDIS_URL")

    if redis_url and SELF_SIGNED_CERTS:
        # https://stackoverflow.com/q/65042551/408556
        redis_url += "?ssl_cert_reqs=none"

    if redis_url and db:
        config["CACHE_REDIS_URL"] = "{}/{}".format(redis_url, db)

    options = {k: v for k, v in kwargs.items() if v is not None}
    options.setdefault("CACHE_THRESHOLD", DEF_THRESHOLD)
    options.setdefault("CACHE_DEFAULT_TIMEOUT", DEF_DEFAULT_TIMEOUT)
    config.update(options)
    return config


def get_pylibmc_client(servers, timeout=None, binary=True, **kwargs):
    from pylibmc import Client

    try:
        from pylibmc import TooBig
    except ImportError:
        from pylibmc import Error, ServerError

        TooBig = (Error, ServerError)

    if timeout:
        kwargs["behaviors"] = {"connect_timeout": timeout}

    client = Client(servers, binary=binary, **kwargs)
    client.TooBig = TooBig
    return client


def get_pymemcache_client(servers, timeout=None, **kwargs):
    from pymemcache.client.hash import HashClient
    from pymemcache.serde import (
        python_memcache_deserializer,
        python_memcache_serializer,
    )

    kwargs.setdefault("serializer", python_memcache_serializer)
    kwargs.setdefault("deserializer", python_memcache_deserializer)

    if timeout:
        kwargs["timeout"] = timeout

    split = [s.split(":") for s in servers]
    _servers = [(host, int(port)) for host, port in split]
    client = HashClient(_servers, **kwargs)

    try:
        client.TooBig = ConnectionResetError
    except NameError:
        import socket

        client.TooBig = socket.error

    return client


def get_bmemcached_client(servers, timeout=None, **kwargs):
    from bmemcached import Client

    if timeout:
        kwargs["socket_timeout"] = timeout

    client = Client(servers, **kwargs)
    client.TooBig = None
    return client


def decode(word):
    try:
        decoded = word.decode(ENCODING)
    except AttributeError:
        decoded = word

    return decoded


def hashable(obj):
    try:
        items = frozenset((k, hashable(v)) for k, v in obj.iteritems())
    except TypeError:
        items = tuple(hashable(item) for item in obj)
    except AttributeError:
        items = obj

    return items


# https://stackoverflow.com/questions/11815873
def simple_memoize(func):
    seen = {}

    @wraps(func)
    def wrapper(*args, **kwargs):
        hashed = hashable(args[0])

        if hashed in seen:
            new_piece = seen[hashed]
        else:
            new_piece = func(*args, **kwargs)
            seen.update({hashed: new_piece})

        return new_piece

    return wrapper



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# vim: sw=4:ts=4:expandtab
"""
manage.py
~~~~~~~~~

A script to manage development tasks
"""

from os import environ
from pathlib import Path
from subprocess import CalledProcessError, call, check_call

import click

from mezmorize.utils import parse_verbosity

BASEDIR = Path(__file__).parent
# DEF_WHERE = ["app", "manage.py", "config.py"]


@click.group()
@click.version_option()
@click.option(
    "-v",
    "--verbose",
    help="Specify multiple times to increase logging verbosity (overridden by -q)",
    count=True,
)
@click.option("-q", "--quiet", help="Only log errors (overrides -v)", is_flag=True)
@click.pass_context
def manager(ctx, verbose=0, quiet=False):
    """CLI manager."""
    verbose = ctx.params["verbose"]
    environ["VERBOSITY"] = parse_verbosity(verbose, quiet)


def upload_():
    """Upload distribution files"""
    upload_dir = BASEDIR.joinpath("dist", "*")
    url = "https://upload.pypi.org/legacy/"
    check_call(f"twine upload --repository-url {url} {upload_dir}", shell=True)


def build_():
    """Create a source distribution package"""
    check_call(BASEDIR.joinpath("helpers", "build"))


def clean_():
    """Remove Python file and build artifacts"""
    check_call(BASEDIR.joinpath("helpers", "clean"))


@manager.command()
def check():
    """Check staged changes for lint errors"""
    exit(call(BASEDIR.joinpath("helpers", "check-stage")))


@manager.command()
@click.option("-w", "--where", help="Modules to check")
@click.option("-f", "--fix", help="Fix errors", is_flag=True)
@click.option("-s", "--strict", help="Check with pylint", is_flag=True)
@click.option(
    "-p",
    "--parallel",
    help="Run linter in parallel in multiple processes",
    is_flag=True,
)
def lint(where=None, fix=False, strict=False, parallel=False):
    """Check style with linters"""
    args = "pylint --rcfile=tests/.pylintrc -rn mezmorize"
    args += " -j 0" if parallel else ""
    r_args = "ruff check --fix" if fix else "ruff check"
    r_args += f" {where}" if where else ""

    try:
        check_call(r_args.split(" "))
        check_call(args.split(" ")) if strict else None
    except CalledProcessError as e:
        exit(e.returncode)


@manager.command()
@click.option("-w", "--where", help="Modules to check")
@click.option("-s", "--sort", help="Sort module imports", is_flag=True)
def prettify(where, sort=False):
    """Prettify code with ruff"""
    extra = where.split(" ") if where else []

    try:
        args = ["ruff", "check", "--select", "I", "--fix"]
        check_call(args + extra) if sort else None
        check_call(["ruff", "format"] + extra)
    except CalledProcessError as e:
        exit(e.returncode)


@manager.command()
def pipme():
    """Install requirements.txt"""
    exit(call("pip install -r requirements.txt".split(" ")))


@manager.command()
def require():
    """Create requirements.txt"""
    cmd = "pip freeze -l | grep -vxFf dev-requirements.txt > requirements.txt"
    exit(call(cmd, shell=True))


@manager.command()
@click.option("-s", "--source", help="the tests to run", default=None)
@click.option("-x", "--stop", help="Stop after first error", is_flag=True)
@click.option("-f", "--failed", help="Run failed tests", is_flag=True)
@click.option("-c", "--cover", help="Add coverage report", is_flag=True)
@click.option("-t", "--tox", help="Run tox tests", is_flag=True)
@click.option("-v", "--verbose", help="Use detailed errors", is_flag=True)
@click.option(
    "-p", "--parallel", help="Run tests in parallel in multiple processes", is_flag=True
)
def test(source=None, where=None, stop=False, **kwargs):
    """Run pytest, tox, and script tests"""
    opts = "-xv" if stop else "-v"
    opts += " --cov=mezmorize" if kwargs.get("cover") else ""
    opts += " --last-failed" if kwargs.get("failed") else ""
    opts += " --numprocesses=auto" if kwargs.get("parallel") else ""
    opts += " --tb=long -ra" if kwargs.get("verbose") else ""
    opts += f" {source}" if source else ""

    try:
        if kwargs.get("tox"):
            check_call(["tox", "-p"] if kwargs.get("parallel") else "tox")
        else:
            check_call(("pytest %s" % opts).split(" "))
    except CalledProcessError as e:
        exit(e.returncode)


@manager.command()
def register():
    """Register package with PyPI"""
    exit(call(["python", BASEDIR.joinpath("setup.py"), "register"]))


@manager.command()
def release():
    """Package and upload a release"""
    try:
        clean_()
        build_()
        upload_()
    except CalledProcessError as e:
        exit(e.returncode)


@manager.command()
def build():
    """Create a source distribution and wheel package"""
    try:
        clean_()
        build_()
    except CalledProcessError as e:
        exit(e.returncode)


@manager.command()
def upload():
    """Upload distribution files"""
    try:
        upload_()
    except CalledProcessError as e:
        exit(e.returncode)


@manager.command()
def clean():
    """Remove Python file and build artifacts"""
    try:
        clean_()
    except CalledProcessError as e:
        exit(e.returncode)


if __name__ == "__main__":
    manager()



# -*- coding: utf-8 -*-
# vim: sw=4:ts=4:expandtab
"""
tests.test_cache
~~~~~~~~~~~~~~~~

Provides unit tests.
"""

import random
import time

import pytest

from mezmorize import Cache, function_namespace
from mezmorize.backends import (
    AVAIL_MEMCACHES,
    FileSystemCache,
    MemcachedCache,
    RedisCache,
    SASLMemcachedCache,
    SimpleCache,
    SpreadSASLMemcachedCache,
)
from mezmorize.utils import HAS_MEMCACHE, HAS_REDIS, get_cache_config

BIGINT = 2**21
BIGGERINT = 2**28


def setup_func(*args, **kwargs):
    namespace = kwargs.pop("namespace", None)
    client_name = kwargs.pop("client_name", None)

    if client_name:
        CACHE_OPTIONS = kwargs.get("CACHE_OPTIONS", {})
        CACHE_OPTIONS["preferred_memcache"] = client_name
        kwargs["CACHE_OPTIONS"] = CACHE_OPTIONS

    config = get_cache_config(*args, **kwargs)
    cache = Cache(namespace=namespace, **config)
    return cache


def check_cache_type(cache, cache_type):
    assert cache.config["CACHE_TYPE"] == cache_type


def check_cache_instance(cache, cache_instance):
    assert isinstance(cache.cache, cache_instance)


def check_client_name(cache, expected):
    assert cache.cache.client_name == expected


def check_too_big(cache, times, error=None):
    if cache.cache.TooBig:
        with pytest.raises(error or cache.cache.TooBig):
            cache.set("big", "a" * times)

        assert cache.get("big") is None
    else:
        cache.set("big", "a" * times)
        assert cache.get("big") == "a" * times


def check_set_delete(cache, key, value, multiplier=None):
    if multiplier:
        value *= multiplier

    cache.set(key, value)
    assert cache.get(key) == value

    cache.delete(key)
    assert cache.get(key) is None


class CacheClient:
    def __init__(self, cache_type="simple", **kwargs):
        self.cache = setup_func(cache_type, **kwargs)

    def clear_cache(self):
        self.cache.clear()


@pytest.fixture
def simple_client():
    client = CacheClient()
    yield client
    client.clear_cache()


@pytest.fixture
def timeout_client():
    client = CacheClient(CACHE_DEFAULT_TIMEOUT=1)
    yield client
    client.clear_cache()


@pytest.fixture
def namespace_client():
    client = CacheClient(namespace="https://github.com/reubano/mezmorize")
    yield client
    client.clear_cache()


@pytest.fixture
def fs_client():
    client = CacheClient("filesystem", CACHE_DIR="/tmp")
    yield client
    client.clear_cache()


@pytest.fixture(params=AVAIL_MEMCACHES)
def memcached_client(request):
    client = CacheClient("memcached", client_name=request.param)
    yield client
    client.clear_cache()


@pytest.fixture(params=AVAIL_MEMCACHES)
def saslmemcached_client(request):
    client = CacheClient("saslmemcached", client_name=request.param)
    yield client
    client.clear_cache()


@pytest.fixture(params=AVAIL_MEMCACHES)
def spreadsaslmemcached_client(request):
    client = CacheClient("spreadsaslmemcached", client_name=request.param)
    yield client
    client.clear_cache()


@pytest.fixture
def redis_client():
    client = CacheClient("redis", db=0)
    yield client
    client.clear_cache()


@pytest.fixture
def redis_custom_client():
    client = CacheClient("redis", db=2)
    yield client
    client.clear_cache()


class TestCache:
    def test_dict_config(self, simple_client):
        check_cache_type(simple_client.cache, "simple")
        check_cache_instance(simple_client.cache, SimpleCache)

    def test_000_set(self, simple_client):
        simple_client.cache.set("hi", "hello")
        assert simple_client.cache.get("hi") == "hello"

    def test_add(self, simple_client):
        simple_client.cache.add("hi", "hello")
        assert simple_client.cache.get("hi") == "hello"

        simple_client.cache.add("hi", "foobar")
        assert simple_client.cache.get("hi") == "hello"

    def test_add_unicode(self, simple_client):
        simple_client.cache.add("ȟį", "ƕɛĺłö")
        assert simple_client.cache.get("ȟį") == "ƕɛĺłö"

        simple_client.cache.add("ȟį", "fööƀåř")
        assert simple_client.cache.get("ȟį") == "ƕɛĺłö"

    def test_add_bytes(self, simple_client):
        simple_client.cache.add(b"hi", b"hello")
        assert simple_client.cache.get(b"hi") == b"hello"

        simple_client.cache.add(b"hi", b"foobar")
        assert simple_client.cache.get(b"hi") == b"hello"

    def test_delete(self, simple_client):
        check_set_delete(simple_client.cache, "hi", "hello")

    def test_delete_unicode(self, simple_client):
        check_set_delete(simple_client.cache, "ȟį", "ƕɛĺłö")

    def test_delete_bytes(self, simple_client):
        check_set_delete(simple_client.cache, b"hi", b"foobar")

    def test_memoize(self, simple_client):
        @simple_client.cache.memoize(5)
        def func(a, b):
            return a + b + random.randrange(0, 100000)

        result = func(5, 2)
        time.sleep(1)
        assert func(5, 2) == result

        result2 = func(5, 3)
        assert result2 != result

        time.sleep(6)
        assert func(5, 2) != result

        time.sleep(1)
        assert func(5, 3) != result2

    def test_timeout(self, timeout_client):
        @timeout_client.cache.memoize(50)
        def func(a, b):
            return a + b + random.randrange(0, 100000)

        result = func(5, 2)
        time.sleep(2)
        assert func(5, 2) == result

    def test_delete_timeout(self, simple_client):
        @simple_client.cache.memoize(5)
        def func(a, b):
            return a + b + random.randrange(0, 100000)

        result = func(5, 2)
        result2 = func(5, 3)
        time.sleep(1)

        assert func(5, 2) == result
        assert func(5, 2) == result
        assert func(5, 3) != result
        assert func(5, 3) == result2

        simple_client.cache.delete_memoized(func)
        assert func(5, 2) != result
        assert func(5, 3) != result2

    def test_delete_verhash(self, simple_client):
        @simple_client.cache.memoize(5)
        def func(a, b):
            return a + b + random.randrange(0, 100000)

        result = func(5, 2)
        result2 = func(5, 3)
        time.sleep(1)

        assert func(5, 2) == result
        assert func(5, 2) == result
        assert func(5, 3) != result
        assert func(5, 3) == result2

        fname = function_namespace(func)[0]
        version_key = simple_client.cache._memvname(fname)
        assert simple_client.cache.get(version_key) is not None

        simple_client.cache.delete_memoized_verhash(func)
        assert simple_client.cache.get(version_key) is None
        assert func(5, 2) != result
        assert func(5, 3) != result2
        assert simple_client.cache.get(version_key) is not None

    def test_delete_rand(self, simple_client):
        @simple_client.cache.memoize()
        def func(a, b):
            return a + b + random.randrange(0, 100000)

        result_a = func(5, 1)
        result_b = func(5, 2)
        assert func(5, 1) == result_a
        assert func(5, 2) == result_b

        simple_client.cache.delete_memoized(func, 5, 2)
        assert func(5, 1) == result_a
        assert func(5, 2) != result_b

    def test_args(self, simple_client):
        @simple_client.cache.memoize()
        def func(a, b):
            return sum(a) + sum(b) + random.randrange(0, 100000)

        result_a = func([5, 3, 2], [1])
        result_b = func([3, 3], [3, 1])
        assert func([5, 3, 2], [1]) == result_a
        assert func([3, 3], [3, 1]) == result_b

        simple_client.cache.delete_memoized(func, [5, 3, 2], [1])
        assert func([5, 3, 2], [1]) != result_a
        assert func([3, 3], [3, 1]) == result_b

    def test_kwargs(self, simple_client):
        @simple_client.cache.memoize()
        def func(a, b=None):
            return a + sum(b.values()) + random.randrange(0, 100000)

        result_a = func(1, {"one": 1, "two": 2})
        result_b = func(5, {"three": 3, "four": 4})
        assert func(1, {"one": 1, "two": 2}) == result_a
        assert func(5, {"three": 3, "four": 4}) == result_b

        simple_client.cache.delete_memoized(func, 1, {"one": 1, "two": 2})
        assert func(1, {"one": 1, "two": 2}) != result_a
        assert func(5, {"three": 3, "four": 4}) == result_b

    def test_kwargonly(self, simple_client):
        @simple_client.cache.memoize()
        def func(a=None):
            if a is None:
                a = 0
            return a + random.random()

        result_a = func()
        result_b = func(5)

        assert func() == result_a
        assert func() < 1
        assert func(5) == result_b
        assert func(5) >= 5
        assert func(5) < 6

    def test_arg_kwarg(self, simple_client):
        @simple_client.cache.memoize()
        def func(a, b, c=1):
            return a + b + c + random.randrange(0, 100000)

        assert func(1, 2) == func(1, 2, c=1)
        assert func(1, 2) == func(1, 2, 1)
        assert func(1, 2) == func(1, 2)
        assert func(1, 2, 3) != func(1, 2)

        with pytest.raises(TypeError):
            func(1)

    def test_classarg(self, simple_client):
        @simple_client.cache.memoize()
        def func(a):
            return a.value + random.random()

        class Adder(object):
            def __init__(self, value):
                self.value = value

        adder = Adder(15)
        adder2 = Adder(20)

        y = func(adder)
        z = func(adder2)
        assert y != z
        assert func(adder) == y
        assert func(adder) != z

        adder.value = 14
        assert func(adder) == y
        assert func(adder) != z
        assert func(adder) != func(adder2)
        assert func(adder2) == z

    def test_classfunc(self, simple_client):
        class Adder(object):
            def __init__(self, initial):
                self.initial = initial

            @simple_client.cache.memoize()
            def add(self, b):
                return self.initial + b

        adder1 = Adder(1)
        adder2 = Adder(2)

        x = adder1.add(3)
        assert adder1.add(3) == x
        assert adder1.add(4) != x
        assert adder1.add(3) != adder2.add(3)

    def test_delete_classfunc(self, simple_client):
        class Adder(object):
            def __init__(self, initial):
                self.initial = initial

            @simple_client.cache.memoize()
            def add(self, b):
                return self.initial + b + random.random()

        adder1 = Adder(1)
        adder2 = Adder(2)

        a1 = adder1.add(3)
        a2 = adder2.add(3)
        assert a1 != a2
        assert adder1.add(3) == a1
        assert adder2.add(3) == a2

        simple_client.cache.delete_memoized(adder1.add)
        a3 = adder1.add(3)
        a4 = adder2.add(3)

        assert a1 != a3
        assert a2 != a3
        assert a2 == a4

        simple_client.cache.delete_memoized(Adder.add)
        a5 = adder1.add(3)
        a6 = adder2.add(3)

        assert a3 != a5
        assert a4 != a6
        assert a5 != a6

    def test_delete_classmethod(self, simple_client):
        class Mock(object):
            @classmethod
            @simple_client.cache.memoize(5)
            def func(cls, a, b):
                return a + b + random.randrange(0, 100000)

        result = Mock.func(5, 2)
        result2 = Mock.func(5, 3)
        time.sleep(1)

        assert Mock.func(5, 2) == result
        assert Mock.func(5, 2) == result
        assert Mock.func(5, 3) != result
        assert Mock.func(5, 3) == result2

        simple_client.cache.delete_memoized(Mock.func)
        assert Mock.func(5, 2) != result
        assert Mock.func(5, 3) != result2

    def test_multiple_arg_kwarg_calls(self, simple_client):
        @simple_client.cache.memoize()
        def func(a, b, c=[1, 1], d=[1, 1]):
            rand = random.randrange(0, 100000)
            return sum(a) + sum(b) + sum(c) + sum(d) + rand

        expected = func([5, 3, 2], [1], c=[3, 3], d=[3, 3])
        assert func([5, 3, 2], [1], d=[3, 3], c=[3, 3]) == expected

        result = func(b=[1], a=[5, 3, 2], c=[3, 3], d=[3, 3])
        assert result == expected
        assert func([5, 3, 2], [1], [3, 3], [3, 3]) == expected

    def test_delete_multiple_arg_kwarg(self, simple_client):
        @simple_client.cache.memoize()
        def func(a, b, c=[1, 1], d=[1, 1]):
            rand = random.randrange(0, 100000)
            return sum(a) + sum(b) + sum(c) + sum(d) + rand

        result_a = func([5, 3, 2], [1], c=[3, 3], d=[3, 3])
        simple_client.cache.delete_memoized(func, [5, 3, 2], [1], [3, 3], [3, 3])
        result_b = func([5, 3, 2], [1], c=[3, 3], d=[3, 3])
        assert result_a != result_b

        simple_client.cache.delete_memoized(func, [5, 3, 2], b=[1], c=[3, 3], d=[3, 3])
        result_b = func([5, 3, 2], [1], c=[3, 3], d=[3, 3])
        assert result_a != result_b

        simple_client.cache.delete_memoized(func, [5, 3, 2], [1], c=[3, 3], d=[3, 3])
        result_a = func([5, 3, 2], [1], c=[3, 3], d=[3, 3])
        assert result_a != result_b

        simple_client.cache.delete_memoized(func, [5, 3, 2], b=[1], c=[3, 3], d=[3, 3])
        result_a = func([5, 3, 2], [1], c=[3, 3], d=[3, 3])
        assert result_a != result_b

        simple_client.cache.delete_memoized(func, [5, 3, 2], [1], c=[3, 3], d=[3, 3])
        result_b = func([5, 3, 2], [1], c=[3, 3], d=[3, 3])
        assert result_a != result_b

        simple_client.cache.delete_memoized(func, [5, 3, 2], [1], [3, 3], [3, 3])
        result_a = func([5, 3, 2], [1], c=[3, 3], d=[3, 3])
        assert result_a != result_b

    def test_kwargs_to_args(self, simple_client):
        def func(a, b, c=None, d=None):
            return sum(a) + sum(b) + random.randrange(0, 100000)

        expected = (1, 2, "foo", "bar")

        args = simple_client.cache._gen_args(func, 1, 2, "foo", "bar")
        assert tuple(args) == expected

        args = simple_client.cache._gen_args(func, 2, "foo", "bar", a=1)
        assert tuple(args) == expected

        args = simple_client.cache._gen_args(func, a=1, b=2, c="foo", d="bar")
        assert tuple(args) == expected

        args = simple_client.cache._gen_args(func, d="bar", b=2, a=1, c="foo")
        assert tuple(args) == expected

        args = simple_client.cache._gen_args(func, 1, 2, d="bar", c="foo")
        assert tuple(args) == expected


class TestNSCache(object):
    def test_memoize(self, namespace_client):
        def func(a, b):
            return a + b + random.randrange(0, 100000)

        cache = setup_func("simple", namespace=namespace_client.cache.namespace)
        cache_key1 = namespace_client.cache._memoize_make_cache_key()(func)
        cache_key2 = cache._memoize_make_cache_key()(func)
        assert cache_key1 == cache_key2


class TestFileSystemCache(TestCache):
    def test_dict_config(self, fs_client):
        check_cache_type(fs_client.cache, "filesystem")
        check_cache_instance(fs_client.cache, FileSystemCache)

    def test_add(self, fs_client):
        fs_client.cache.add("hi", "hello")
        assert fs_client.cache.get("hi") == "hello"

        fs_client.cache.add("hi", "foobar")
        assert fs_client.cache.get("hi") == "hello"

    def test_add_unicode(self, fs_client):
        fs_client.cache.add("ȟį", "ƕɛĺłö")
        assert fs_client.cache.get("ȟį") == "ƕɛĺłö"

        fs_client.cache.add("ȟį", "fööƀåř")
        assert fs_client.cache.get("ȟį") == "ƕɛĺłö"


if HAS_MEMCACHE:

    class TestMemcachedCache(TestCache):
        def test_dict_config(self, memcached_client):
            check_cache_type(memcached_client.cache, "memcached")
            check_cache_instance(memcached_client.cache, MemcachedCache)
            # check_client_name(memcached_client.cache, client_name)

        def test_mc_large_value(self, memcached_client):
            check_too_big(memcached_client.cache, BIGGERINT)
else:
    print("TestMemcachedCache requires Memcache")

if HAS_MEMCACHE:

    class TestSASLMemcachedCache(TestCache):
        def test_dict_config(self, saslmemcached_client):
            check_cache_type(saslmemcached_client.cache, "saslmemcached")
            check_cache_instance(saslmemcached_client.cache, SASLMemcachedCache)

        def test_mc_large_value(self, saslmemcached_client):
            check_too_big(saslmemcached_client.cache, BIGGERINT)

else:
    print("TestSASLMemcachedCache requires Memcache")

if HAS_MEMCACHE:

    class TestSpreadSASLMemcachedCache(TestCache):
        def test_dict_config(self, spreadsaslmemcached_client):
            cache_instance = SpreadSASLMemcachedCache
            check_cache_type(spreadsaslmemcached_client.cache, "spreadsaslmemcached")
            check_cache_instance(spreadsaslmemcached_client.cache, cache_instance)

        def test_mc_large_value(self, spreadsaslmemcached_client):
            check_set_delete(spreadsaslmemcached_client.cache, "big", "a", BIGINT)
            check_set_delete(spreadsaslmemcached_client.cache, "ƅıɠ", "ą", BIGINT)
            check_set_delete(spreadsaslmemcached_client.cache, b"big", b"a", BIGINT)
            check_too_big(spreadsaslmemcached_client.cache, BIGGERINT, ValueError)
else:
    print("TestSpreadSASLMemcachedCache requires Memcache")

if HAS_REDIS:

    class TestRedisCache(TestCache):
        def test_dict_config(self, redis_client):
            check_cache_type(redis_client.cache, "redis")
            check_cache_instance(redis_client.cache, RedisCache)

        def test_add_bytes(self, redis_client):
            redis_client.cache.add(b"hi", b"hello")
            assert redis_client.cache.get(b"hi") == b"hello"

            redis_client.cache.add(b"hi", b"foobar")
            assert redis_client.cache.get(b"hi") == b"hello"

        def test_delete_bytes(self, redis_client):
            check_set_delete(redis_client.cache, b"hi", b"hello")

        def test_redis_url_default_db(self, redis_client):
            client = redis_client.cache.cache._client
            rconn = client.connection_pool.get_connection("foo")
            assert rconn.db == 0

        def test_redis_url_custom_db(self, redis_custom_client):
            client = redis_client.cache.cache._client
            rconn = client.connection_pool.get_connection("foo")
            assert rconn.db == 2
else:
    print("TestRedisCache requires Redis")

if __name__ == "__main__":
    pytest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# vim: sw=4:ts=4:expandtab
"""
mezmorize.backends
~~~~~~~~~~~~~~~~~~

Provides mezmorize storage mechanisms
"""
# pylint: disable=range-builtin-not-iterating,filter-builtin-not-iterating

import pickle
from functools import partial
from itertools import chain
from operator import contains

from cachelib import (
    FileSystemCache,
    NullCache,
    RedisCache,
    SimpleCache,
)
from cachelib import (
    MemcachedCache as _MemcachedCache,
)

from .utils import (
    AVAIL_MEMCACHES,
    DEF_MC_SERVERS,
    DEF_REDIS_HOST,
    DEF_REDIS_PORT,
    HAS_MEMCACHE,
    get_bmemcached_client,
    get_pylibmc_client,
    get_pymemcache_client,
)

try:
    from redis import from_url
except ImportError:
    from_url = None

CONFIG_LOOKUP = {
    "servers": "CACHE_MEMCACHED_SERVERS",
    "threshold": "CACHE_THRESHOLD",
    "username": "CACHE_MEMCACHED_USERNAME",
    "password": "CACHE_MEMCACHED_PASSWORD",
    "key_prefix": "CACHE_KEY_PREFIX",
    "timeout": "connect_timeout",
}


def gen_defaults(*keys, **config):
    for key in keys:
        config_key = CONFIG_LOOKUP[key]

        if config_key in config:
            yield (key, config[config_key])


def get_mc_client(module_name, binary=True, **kwargs):
    servers = kwargs.pop("servers", (DEF_MC_SERVERS,))
    timeout = kwargs.pop("timeout", None)

    if module_name == "pylibmc":
        client = get_pylibmc_client(servers, timeout=timeout, binary=binary, **kwargs)
    elif module_name == "pymemcache":
        client = get_pymemcache_client(servers, timeout=timeout, **kwargs)
    elif module_name == "bmemcached":
        client = get_bmemcached_client(servers, timeout=timeout, **kwargs)

    return client


class MemcachedCache(_MemcachedCache):
    def __init__(self, *args, **kwargs):
        if not HAS_MEMCACHE:
            raise RuntimeError("No memcache module found.")

        compat_memcaches = kwargs.pop("compat_memcaches", AVAIL_MEMCACHES)
        avail_memcaches = set(AVAIL_MEMCACHES).intersection(compat_memcaches)

        if not avail_memcaches:
            raise RuntimeError("No compatible memcache module found.")

        preferred_mc = kwargs.get("preferred_memcache", "pylibmc")

        if len(avail_memcaches) == 1 or preferred_mc not in avail_memcaches:
            filterer = partial(contains, avail_memcaches)
            preferred_mc = next(filter(filterer, AVAIL_MEMCACHES))

        whitelist = {"default_timeout", "key_prefix"}
        blacklist = whitelist.union(["preferred_memcache"])
        mkwargs = {k: v for k, v in kwargs.items() if k not in blacklist}
        skwargs = {k: v for k, v in kwargs.items() if k in whitelist}

        client = get_mc_client(preferred_mc, **mkwargs)
        super(MemcachedCache, self).__init__(servers=client, **skwargs)
        self.TooBig = client.TooBig
        self.client_name = preferred_mc


class SASLMemcachedCache(MemcachedCache):
    def __init__(self, *args, **kwargs):
        kwargs["compat_memcaches"] = ("pylibmc", "bmemcached")
        super(SASLMemcachedCache, self).__init__(*args, **kwargs)


def null(config, *args, **kwargs):
    return NullCache()


def simple(config, *args, **kwargs):
    defaults = dict(gen_defaults("threshold", **config))
    defaults.update(kwargs)
    return SimpleCache(*args, **defaults)


def memcached(config, *args, **kwargs):
    keys = ("timeout", "servers", "key_prefix")
    defaults = dict(gen_defaults(*keys, **config))
    defaults.update(kwargs)
    return MemcachedCache(*args, **defaults)


def saslmemcached(config, **kwargs):
    keys = ("timeout", "servers", "username", "password", "key_prefix")
    defaults = dict(gen_defaults(*keys, **config))
    defaults.update(kwargs)
    return SASLMemcachedCache(**defaults)


def filesystem(config, *args, **kwargs):
    args = chain([config["CACHE_DIR"]], args)
    defaults = dict(gen_defaults("threshold", **config))
    defaults.update(kwargs)
    return FileSystemCache(*args, **defaults)


def redis(config, *args, **kwargs):
    kwargs.setdefault("host", config.get("CACHE_REDIS_HOST", DEF_REDIS_HOST))
    kwargs.setdefault("port", config.get("CACHE_REDIS_PORT", DEF_REDIS_PORT))
    kwargs.setdefault("password", config.get("CACHE_REDIS_PASSWORD"))
    kwargs.setdefault("key_prefix", config.get("CACHE_KEY_PREFIX"))
    kwargs.setdefault("db", config.get("CACHE_REDIS_DB"))
    redis_url = config.get("CACHE_REDIS_URL")

    if redis_url:
        kwargs["host"] = from_url(redis_url, db=kwargs.pop("db", None))

    return RedisCache(*args, **kwargs)


class SpreadSASLMemcachedCache(SASLMemcachedCache):
    """
    Simple Subclass of SASLMemcached client that spread value across multiple
    key is they are bigger than a given threshold.

    Spreading require using pickle to store the value, which can significantly
    impact the performances.
    """

    DEF_CHUNKSIZE = 2**20 - 2**7  # 1048448
    DEF_MAXCHUNKS = 32

    def __init__(self, *args, **kwargs):
        """
        Kwargs:
            chunksize (int): max length of a pickled object that can fit in
                memcached (memcache has an upper limit of 1MB for values,
                default: 1048448)
        """
        self.CHUNKSIZE = kwargs.get("chunksize", self.DEF_CHUNKSIZE)
        self.MARKER = "SpreadSASLMemcachedCache.SpreadedValue"
        self.MAXCHUNKS = kwargs.get("maxchunks", self.DEF_MAXCHUNKS)
        super(SpreadSASLMemcachedCache, self).__init__(*args, **kwargs)
        self.super = super(SpreadSASLMemcachedCache, self)

    def _genkeys(self, key):
        return ("{}.{}".format(key, i) for i in range(self.MAXCHUNKS))

    def _gen_kv(self, key, pickled):
        chunks = range(0, len(pickled), self.CHUNKSIZE)

        if len(chunks) > self.MAXCHUNKS:
            msg = "Value exceed maximum number of keys ({})"
            raise ValueError(msg.format(self.MAXCHUNKS))

        for i in chunks:
            _key = "{}.{}".format(key, i // self.CHUNKSIZE)
            _value = pickled[i : i + self.CHUNKSIZE]
            yield _key, _value

    def delete(self, key):
        value = self.super.get(key)
        self.super.delete(key)

        if value == self.MARKER:
            for skey in self._genkeys(key):
                self.super.delete(skey)

    def set(self, key, value, timeout=None):
        """set a value in cache, potentially spreading it across multiple key."""
        try:
            value = self.super.set(key, value, timeout=timeout)
        except self.TooBig:
            self.super.set(key, self.MARKER, timeout=timeout)
            pickled = pickle.dumps(value, 2)
            values = dict(self._gen_kv(key, pickled))
            value = self.super.set_many(values, timeout)

        return value

    def get(self, key):
        """get a value in cache, potentially from multiple keys."""
        value = self.super.get(key)

        if value == self.MARKER:
            keys = self._genkeys(key)
            result = self.super.get_many(*keys)
            serialized = b"".join(v for v in result if v is not None)
            value = pickle.loads(serialized) if serialized else None

        return value


def spreadsaslmemcached(config, *args, **kwargs):
    keys = ("timeout", "servers", "username", "password", "key_prefix")
    defaults = dict(gen_defaults(*keys, **config))
    defaults.update(kwargs)
    return SpreadSASLMemcachedCache(*args, **defaults)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# vim: sw=4:ts=4:expandtab
"""
mezmorize
~~~~~~~~~

Adds function memoization support
"""
# pylint: disable=W1636,W1637,W1638,W1639

import base64
import hashlib
import inspect
import uuid
import warnings
from functools import partial, wraps
from importlib import import_module
from inspect import getfullargspec

from cachelib.memcached import _test_memcached_key

from . import backends
from .utils import (
    DEF_DEFAULT_TIMEOUT,
    DEF_THRESHOLD,
    ENCODING,
    decode,
    get_cache_config,
    get_cache_type,
)

__version__ = "0.30.0"
__copyright__ = "Copyright 2015 Reuben Cummings"


# Used to remove control characters and whitespace from cache keys.
def is_invalid(char):
    return not (char in {"_", "."} or char.isalnum())


delchars = filter(is_invalid, map(chr, range(256)))
trans_tbl = "".maketrans({k: None for k in delchars})
NULL_CONTROL = (trans_tbl,)
FIRST_NC = NULL_CONTROL[0]


def get_namespace(*names):
    text = ".".join(map(decode, names))

    if FIRST_NC:
        namespace = text.translate(FIRST_NC)
    else:
        encoded = text.encode(ENCODING)
        namespace = encoded.translate(*NULL_CONTROL).decode(ENCODING)

    return namespace


def function_namespace(f, *args):
    """
    Attempts to returns unique a namespace for a function
    """
    m_args = getfullargspec(f).args
    m_arg = m_args[0] if args and m_args else ""
    arg = args[0] if args else None
    self_instance = getattr(f, "__self__", None)
    not_class = self_instance and not inspect.isclass(self_instance)
    is_self = m_arg == "self"

    try:
        name = f.__qualname__
    except AttributeError:
        klass = self_instance.__class__ if not_class else self_instance

        if not klass:
            klass = getattr(f, "im_class", None)

        if not klass and is_self:
            klass = arg.__class__
        elif not klass and m_arg == "cls":
            klass = arg

        name = ".".join(n.__name__ for n in (klass, f) if n)

    module = f.__module__
    ns = get_namespace(module, name)

    if not_class or is_self:
        instance = f.__self__ if not_class else arg
        ins = get_namespace(module, name, repr(instance))
    else:
        ins = None

    return ns, ins


class Cache(object):
    """
    This class is used to control the cache objects.
    """

    def __init__(self, namespace=None, **config):
        config.setdefault("CACHE_DEFAULT_TIMEOUT", DEF_DEFAULT_TIMEOUT)
        config.setdefault("CACHE_THRESHOLD", DEF_THRESHOLD)
        config.setdefault("CACHE_KEY_PREFIX", "mezmorize_")
        config.setdefault("CACHE_MEMCACHED_SERVERS", None)
        config.setdefault("CACHE_DIR", None)
        config.setdefault("CACHE_OPTIONS", {})
        config.setdefault("CACHE_ARGS", [])
        config.setdefault("CACHE_TYPE", "simple")
        config.setdefault("CACHE_NO_NULL_WARNING", False)

        warning = not config["CACHE_NO_NULL_WARNING"]

        if config["CACHE_TYPE"] == "null" and warning:
            warnings.warn("CACHE_TYPE is set to null, caching is effectively disabled.")

        self.namespace = str(namespace or "")
        self.config = config
        self._set_cache()

    def _set_cache(self):
        module_string = self.config["CACHE_TYPE"]
        default_timeout = self.config["CACHE_DEFAULT_TIMEOUT"]

        if "." not in module_string:
            try:
                cache_obj = getattr(backends, module_string)
            except AttributeError:
                msg = "{} is not a valid Mezmorize backend"
                raise ImportError(msg.format(module_string))
        else:
            cache_obj = import_module(module_string)

        self.cache_type = cache_obj.__name__
        self.is_memcached = "memcache" in self.cache_type

        args = self.config["CACHE_ARGS"]
        kwargs = self.config["CACHE_OPTIONS"]
        kwargs.setdefault("default_timeout", default_timeout)

        if not self.is_memcached:
            kwargs.pop("preferred_memcache", None)
            kwargs.pop("connect_timeout", None)

        self.cache = cache_obj(self.config, *args, **kwargs)

        try:
            self.client_name = self.cache.client_name
        except AttributeError:
            self.client_name = None

    def _gen_mapping(self, *args):
        for key in args:
            if _test_memcached_key(key):
                encoded_key = self.cache._normalize_key(key)
                yield (encoded_key, key)

    # https://github.com/pallets/werkzeug/pull/1161
    def get_values(self, *args):
        are_strings = (isinstance(key, str) for key in args)
        has_encoded_keys = self.cache.key_prefix or not all(are_strings)
        key_mapping = dict(self._gen_mapping(*args))
        keys = list(key_mapping)
        rv = self.cache._client.get_multi(keys)

        if has_encoded_keys:
            rv = {key_mapping[key]: value for key, value in rv.items()}

        return [rv.get(key) for key in args]

    def get(self, *args, **kwargs):
        "Proxy function for internal cache object."
        return self.cache.get(*args, **kwargs)

    def set(self, *args, **kwargs):
        "Proxy function for internal cache object."
        self.cache.set(*args, **kwargs)

    def add(self, *args, **kwargs):
        "Proxy function for internal cache object."
        self.cache.add(*args, **kwargs)

    def delete(self, *args, **kwargs):
        "Proxy function for internal cache object."
        self.cache.delete(*args, **kwargs)

    def delete_many(self, *args, **kwargs):
        "Proxy function for internal cache object."
        self.cache.delete_many(*args, **kwargs)

    def clear(self):
        "Proxy function for internal cache object."
        self.cache.clear()

    def get_many(self, *args, **kwargs):
        "Proxy function for internal cache object."
        if self.is_memcached:
            values = self.get_values(*args)
        else:
            values = self.cache.get_many(*args, **kwargs)

        return values

    def set_many(self, *args, **kwargs):
        "Proxy function for internal cache object."
        self.cache.set_many(*args, **kwargs)

    def _memvname(self, funcname):
        return funcname + "_memver"

    def _memoize_make_version_hash(self):
        if self.namespace.startswith("http"):
            UUID = uuid.uuid3(uuid.NAMESPACE_URL, self.namespace)
        if self.namespace:
            UUID = uuid.uuid3(uuid.NAMESPACE_DNS, self.namespace)
        else:
            UUID = uuid.uuid4()

        return base64.b64encode(UUID.bytes)[:6].decode(ENCODING)

    def _memoize_version(self, f, *args, **kwargs):
        """
        Updates the hash version associated with a memoized function or method.
        """
        reset = kwargs.pop("reset", None)
        delete = kwargs.pop("delete", None)
        fname, instance_fname = function_namespace(f, *args)
        version_key = self._memvname(fname)

        if instance_fname:
            fetch_keys = [version_key, self._memvname(instance_fname)]
        else:
            fetch_keys = [version_key]

        # Only delete the per-instance version key or per-function version
        # key but not both.
        if delete:
            self.cache.delete_many(fetch_keys[-1])
            return fname, None

        version_data_list = list(self.get_many(*fetch_keys))
        dirty = False

        if version_data_list[0] is None:
            version_data_list[0] = self._memoize_make_version_hash()
            dirty = True

        if instance_fname and version_data_list[1] is None:
            version_data_list[1] = self._memoize_make_version_hash()
            dirty = True

        # Only reset the per-instance version or the per-function version
        # but not both.
        if reset:
            fetch_keys = fetch_keys[-1:]
            version_data_list = [self._memoize_make_version_hash()]
            dirty = True

        if dirty:
            zipped = zip(fetch_keys, version_data_list)
            self.cache.set_many(dict(zipped), **kwargs)

        return fname, "".join(map(decode, version_data_list))

    def _memoize_make_cache_key(self, make_name=None, decorated=None):
        """
        Function used to create the cache_key for memoized functions.
        """

        def make_cache_key(f, *args, **kwargs):
            mkwargs = {"timeout": decorated.cache_timeout} if decorated else {}
            fname, version_data = self._memoize_version(f, *args, **mkwargs)

            # this should have to be after version_data, so that it
            # does not break the delete_memoized functionality.
            altfname = make_name(fname) if callable(make_name) else fname

            if callable(f):
                keyargs = tuple(self._gen_args(f, *args, **kwargs))
                keykwargs = {}
            else:
                keyargs, keykwargs = args, kwargs

            updated = "{0}{1}{2}".format(altfname, keyargs, keykwargs)
            cache_key = hashlib.md5()
            cache_key.update(updated.encode(ENCODING))
            cache_key = base64.b64encode(cache_key.digest())[:16]
            cache_key = cache_key.decode(ENCODING)
            cache_key += version_data

            return cache_key

        return make_cache_key

    def _gen_args(self, f, *args, **kwargs):
        # Inspect the arguments to the function
        # This allows the memoization to be the same
        # whether the function was called with
        # 1, b=2 is equivalent to a=1, b=2, etc.
        num_args = len(args)
        argspec = getfullargspec(f)
        _defaults = argspec.defaults or []
        m_args = argspec.args
        defaults = dict(zip(reversed(m_args), reversed(_defaults)))
        counter = 0

        for i, m_arg in enumerate(m_args):
            # Subtract from i, m_args that aren't in args
            arg_num = i - counter

            if not i and m_arg in ("self", "cls"):
                # supports instance methods for the memoized functions
                new_arg = repr(args[0])
            elif kwargs.get(m_arg) is not None:
                new_arg = kwargs[m_arg]
                counter += 1
            elif arg_num < num_args:
                new_arg = args[arg_num]
            elif defaults.get(m_arg) is not None:
                new_arg = defaults[m_arg]
            else:
                new_arg = None

            yield new_arg

    def memoize(self, timeout=None, make_name=None, unless=None):
        """
        Use this to cache the result of a function, taking its arguments into
        account in the cache key.

        Information on
        `Memoization <http://en.wikipedia.org/wiki/Memoization>`_.

        Example::
            >>> import random
            >>>
            >>> cache = Cache()
            >>>
            >>> @cache.memoize(timeout=50)
            ... def big_foo(a, b):
            ...     return a + b + random.random()

        .. code-block:: python

            >>> rand_res = big_foo(5, 2)
            >>> rand_res == big_foo(5, 3)
            False
            >>> rand_res == big_foo(5, 2)
            True

        .. versionadded:: 0.4
            The returned decorated function now has three function attributes
            assigned to it.

                **uncached**
                    The original undecorated function. readable only

                **cache_timeout**
                    The cache timeout value for this function. For a custom
                    value to take affect, this must be set before the function
                    is called.

                    readable and writable

                **make_cache_key**
                    A function used in generating the cache_key used.

                    readable and writable


        :param timeout: Default None. If set to an integer, will cache for that
                        amount of time. Unit of time is in seconds.
        :param make_name: Default None. If set this is a function that accepts
                          a single argument, the function name, and returns a
                          new string to be used as the function name. If not
                          set then the function name is used.
        :param unless: Default None. Cache will *always* execute the caching
                       facilities unless this callable is true.
                       This will bypass the caching entirely.

        .. versionadded:: 0.5
            params ``make_name``, ``unless``
        """

        def _memoize(f):
            @wraps(f)
            def decorated(*args, **kwargs):
                if callable(unless) and unless():  # bypass cache
                    return f(*args, **kwargs)

                cache_key = decorated.make_cache_key(f, *args, **kwargs)
                value = self.cache.get(cache_key)

                if value is None:
                    value = f(*args, **kwargs)
                    ckwargs = {"timeout": decorated.cache_timeout}

                    # value is first for addCallback compatibility
                    def set_cache(value, key):
                        self.cache.set(key, value, **ckwargs)
                        return value

                    try:
                        value.addCallback(set_cache, cache_key)
                    except AttributeError:
                        set_cache(value, cache_key)

                return value

            decorated.uncached = f
            decorated.cache_timeout = timeout
            m_make_cache_key = self._memoize_make_cache_key
            decorated.make_cache_key = m_make_cache_key(make_name, decorated)
            decorated.delete_memoized = partial(self.delete_memoized, f)
            return decorated

        return _memoize

    def delete_memoized(self, f, *args, **kwargs):
        """
        Deletes the specified functions caches, based by given parameters.
        If parameters are given, only the functions that were memoized with
        them will be erased. Otherwise all versions of the caches will be
        forgotten.

        Example::
            >>> import random
            >>>
            >>> cache = Cache()
            >>>
            >>> @cache.memoize(50)
            ... def random_func():
            ...    return random.random()

            >>> @cache.memoize()
            ... def param_func(a, b):
            ...    return a + b + random.random()

        .. code-block:: python

            >>> rand_res = random_func()
            >>> rand_res == random_func()
            True
            >>> cache.delete_memoized(random_func)
            >>> rand_res == random_func()
            False
            >>> param_res = param_func(1, 2)
            >>> param_res == param_func(1, 2)
            True
            >>> param_res2 = param_func(2, 2)
            >>> param_res == param_res2
            False
            >>> cache.delete_memoized(param_func, 1, 2)
            >>> param_res == param_func(1, 2)
            False
            >>> param_res2 == param_func(2, 2)
            True

        Delete memoized is also smart about instance methods vs class methods.

        When passing an instancemethod, it will only clear the cache related
        to that instance of that object. (object uniqueness can be overridden
            by defining the __repr__ method, such as user id).

        When passing a classmethod, it will clear all caches related across
        all instances of that class.

        Example::

            >>> class Adder(object):
            ...    @cache.memoize()
            ...    def add(self, b):
            ...        return b + random.random()

        .. code-block:: python

            >>> adder1 = Adder()
            >>> adder2 = Adder()
            >>> adder1_res = adder1.add(3)
            >>> adder2_res = adder2.add(3)
            >>> cache.delete_memoized(adder1.add)
            >>> adder1_res == adder1.add(3)
            False
            >>> adder1_res = adder1.add(3)
            >>> adder2_res == adder2.add(3)
            True
            >>> cache.delete_memoized(Adder.add)
            >>> adder1_res == adder1.add(3)
            False
            >>> adder2_res == adder2.add(3)
            False

        :param fname: Name of the memoized function, or a reference to
            the function.
        :param *args: A list of positional parameters used with memoized
            function.
        :param **kwargs: A dict of named parameters used with memoized
            function.

        .. note::

            Flask-Cache uses inspect to order kwargs into positional args when
            the function is memoized. If you pass a function reference into
            ``fname`` instead of the function name, Flask-Cache will be able to
            place the args/kwargs in the proper order, and delete the
            positional cache.

            However, if ``delete_memoized`` is just called with the name of the
            function, be sure to pass in potential arguments in the same order
            as defined in your function as args only, otherwise Flask-Cache
            will not be able to compute the same cache key.

        .. note::

            Flask-Cache maintains an internal random version hash for the
            function. Using delete_memoized will only swap out the version
            hash, causing the memoize function to recompute results and put
            them into another key.

            This leaves any computed caches for this memoized function within
            the caching backend.

            It is recommended to use a very high timeout with memoize if using
            this function, so that when the version has been swapped, the old
            cached results would eventually be reclaimed by the caching
            backend.
        """
        if not callable(f):
            raise DeprecationWarning(
                "Deleting messages by relative name is no longer"
                " reliable, please switch to a function reference"
            )

        if not (args or kwargs):
            self._memoize_version(f, reset=True)
        else:
            cache_key = f.make_cache_key(f.uncached, *args, **kwargs)
            self.cache.delete(cache_key)

    def delete_memoized_verhash(self, f, *args):
        """
        Delete the version hash associated with the function.

        ..warning::

            Performing this operation could leave keys behind that have
            been created with this version hash. It is up to the application
            to make sure that all keys that may have been created with this
            version hash at least have timeouts so they will not sit orphaned
            in the cache backend.
        """
        if not callable(f):
            raise DeprecationWarning(
                "Deleting messages by relative name is no longer"
                " reliable, please use a function reference"
            )

        self._memoize_version(f, delete=True)


def get_cache(*args, **kwargs):
    """Helper function to intelligently configure an appropriate cache

    Kwargs:
        cache_type (str): The type of cache backend to use. Default depends on
            installed libraries and running servers.

        spread (bool): Use spreadsaslmemcached if available. Default False.
        preferred_memcache (str): Use this memcached client if available.
            Default None.

        cache_threshold (int): The max number of keys to store.
        cache_options (dict): Passed as kwargs to the cache backend client.
        connect_timeout (int): Max number of seconds to wait for response.
            Default None, e.g., forever.

        cache_default_timeout (int): Number of seconds to store cache result if
            `timeout` is not set.

        self_signed_certs (bool): Allow self-signed certificates.

    Returns:
        decorator: an iterator of items

    Example:
        >>> from os import getenv
        >>> from .utils import HAS_MEMCACHE, HAS_REDIS
        >>>
        >>> cache = get_cache()
        >>> cache.set('key', 'value')
        >>> cache.get('key') == 'value'
        True
        >>> cache.delete('key')
        >>> cache.get('key') is None
        True
        >>>
        >>> if cache.client_name and HAS_MEMCACHE:
        ...     cache.cache_type == 'memcached'
        ...     cache.client_name == 'pylibmc'
        ... elif HAS_REDIS:
        ...     cache.cache_type == 'redis'
        ...     cache.client_name is None
        ... else:
        ...     cache.cache_type == 'filesystem'
        ...     cache.client_name is None
        True
        True
        >>>
        >>> cache = get_cache(cache_type='simple')
        >>> cache.cache_type
        'simple'
        >>>
        >>> cache = get_cache(cache_type='filesystem')
        >>> cache.cache_type
        'filesystem'
        >>>
        >>> cache = get_cache(preferred_memcache='bmemcached')
        >>>
        >>> if cache.client_name:
        ...     cache.client_name == 'bmemcached'
        ... else:
        ...     cache.client_name is None
        True
    """
    _cache_type = kwargs.get("cache_type")
    spread = kwargs.get("spread")
    namespace = kwargs.get("namespace")
    whitelist = {
        "cache_default_timeout",
        "cache_threshold",
        "cache_options",
        "cache_key_prefix",
        "self_signed_certs",
    }

    ckwargs = {k.upper(): v for k, v in kwargs.items() if k in whitelist}

    keys = {"preferred_memcache", "connect_timeout"}
    extra_options = {k: v for k, v in kwargs.items() if k in keys}

    if extra_options:
        CACHE_OPTIONS = ckwargs.get("CACHE_OPTIONS", {})
        CACHE_OPTIONS.update(extra_options)
        ckwargs["CACHE_OPTIONS"] = CACHE_OPTIONS

    cache_type = get_cache_type(cache=_cache_type, spread=spread)
    config = get_cache_config(cache_type, **ckwargs)
    return Cache(namespace=namespace, **config)


def memoize(*args, **kwargs):
    """Use this to cache the result of a function, taking its arguments into
    account in the cache key.

    `Memoization <http://en.wikipedia.org/wiki/Memoization>`_.

    Kwargs:
        timeout (int): Number of seconds to store cache result. Default None,
            e.g., forever. Note: If this is not set, it is overridden by
            `cache_default_timeout`.

        unless (func): Don't use cache if this callable is true. Default None.

    Returns:
        decorator: a memoizer

    Example:
        >>> import random
        >>> from os import getenv
        >>> from .utils import HAS_MEMCACHE, HAS_REDIS
        >>>
        >>> get_rand = lambda: random.random()
        >>> rand_value = get_rand()
        >>> rand_value == get_rand()
        False
        >>> memoizer = memoize(timeout=30)
        >>> memoized_get_rand = memoizer(get_rand)
        >>> memoized_rand_value = memoized_get_rand()
        >>> memoized_rand_value == memoized_get_rand()
        True
        >>>
        >>> if memoizer.client_name and HAS_MEMCACHE:
        ...     memoizer.cache_type == 'memcached'
        ...     memoizer.client_name == 'pylibmc'
        ... elif HAS_REDIS:
        ...     memoizer.cache_type == 'redis'
        ...     memoizer.client_name is None
        ... else:
        ...     memoizer.cache_type == 'filesystem'
        ...     memoizer.client_name is None
        True
        True
        >>>
        >>> memoizer = memoize(cache_type='simple')
        >>> memoizer.cache_type
        'simple'
        >>>
        >>> memoizer = memoize(cache_type='filesystem')
        >>> memoizer.cache_type
        'filesystem'
        >>>
        >>> memoizer = memoize(preferred_memcache='bmemcached')
        >>>
        >>> if memoizer.client_name:
        ...     memoizer.client_name == 'bmemcached'
        ... else:
        ...     memoizer.client_name is None
        True
    """
    cache = get_cache(*args, **kwargs)
    mkwargs = {k: v for k, v in kwargs.items() if k in {"timeout", "unless"}}
    memoizer = cache.memoize(*args, **mkwargs)
    memoizer.cache_type = cache.cache_type
    memoizer.client_name = cache.client_name
    return memoizer
